# app/routers/team.py
from controller.generic import create_crud_router
from model.models import Team 
from model.dto import TeamCreate, TeamUpdate, TeamRead

router = create_crud_router(
    model=Team,
    create_schema=TeamCreate,
    update_schema=TeamUpdate,
    read_schema=TeamRead,
    prefix="/teams",
    tags=["teams"],
)



# app/repositories/base.py
from typing import Generic, Type, TypeVar, Optional, List, Any
from sqlmodel import SQLModel, Session, select

ModelT = TypeVar("ModelT", bound=SQLModel)
CreateT = TypeVar("CreateT", bound=SQLModel)
UpdateT = TypeVar("UpdateT", bound=SQLModel)

class Repository(Generic[ModelT, CreateT, UpdateT]):
    def __init__(self, model: Type[ModelT]):
        self.model = model

    def get(self, session: Session, id: Any) -> Optional[ModelT]:
        return session.get(self.model, id)

    def list(self, session: Session, offset: int = 0, limit: int = 100) -> List[ModelT]:
        stmt = select(self.model).offset(offset).limit(limit)
        return list(session.exec(stmt))

    def create(self, session: Session, data: CreateT) -> ModelT:
        obj = self.model.model_validate(data)  # converte CreateT -> ModelT
        session.add(obj)
        session.commit()
        session.refresh(obj)
        return obj

    def update(self, session: Session, obj: ModelT, data: UpdateT) -> ModelT:
        data_dict = data.model_dump(exclude_unset=True)
        for k, v in data_dict.items():
            setattr(obj, k, v)
        session.add(obj)
        session.commit()
        session.refresh(obj)
        return obj

    def delete(self, session: Session, obj: ModelT) -> None:
        session.delete(obj)
        session.commit()



# app/routers/generic.py
from typing import Any, Callable, Generic, Optional, Type, TypeVar
from fastapi import APIRouter, Depends, HTTPException, Query
from sqlmodel import SQLModel, Session
from util.database import get_session
from repository.base import Repository
from service.base import Service

ModelT = TypeVar("ModelT", bound=SQLModel)
CreateT = TypeVar("CreateT", bound=SQLModel)
UpdateT = TypeVar("UpdateT", bound=SQLModel)
ReadT   = TypeVar("ReadT", bound=SQLModel)

class Hooks(Generic[ModelT, CreateT, UpdateT]):
    """
    Pontos de extensão opcionais para regras específicas.
    """
    def pre_create(self, payload: CreateT, session: Session) -> None: ...
    def pre_update(self, payload: UpdateT, session: Session, obj: ModelT) -> None: ...
    def pre_delete(self, session: Session, obj: ModelT) -> None: ...

def create_crud_router(
    *,
    model: Type[ModelT],
    create_schema: Type[CreateT],
    update_schema: Type[UpdateT],
    read_schema: Type[ReadT],
    prefix: str,
    tags: list[str],
    hooks: Optional[Hooks[ModelT, CreateT, UpdateT]] = None,
    page_size_limit: int = 200,
) -> APIRouter:
    """
    Cria endpoints: POST /, GET /, GET /{id}, PATCH /{id}, DELETE /{id}
    """
    router = APIRouter(prefix=prefix, tags=tags)

    repo: Repository[ModelT, CreateT, UpdateT] = Repository(model)
    service: Service[ModelT, CreateT, UpdateT] = Service(repo)
    _hooks = hooks or Hooks()  # instancia vazia (métodos no-op)

    @router.post("/", response_model=read_schema, status_code=201)
    def create_item(payload: create_schema, session: Session = Depends(get_session)):
        if hasattr(_hooks, "pre_create") and callable(_hooks.pre_create):
            _hooks.pre_create(payload, session)
        return service.create(session, payload)

    @router.get("/", response_model=list[read_schema])
    def list_items(
        session: Session = Depends(get_session),
        offset: int = 0,
        limit: int = Query(100, le=page_size_limit),
    ):
        return service.list(session, offset, limit)

    @router.get("/{item_id}", response_model=read_schema)
    def get_item(item_id: int, session: Session = Depends(get_session)):
        obj = service.get(session, item_id)
        if not obj:
            raise HTTPException(404, "Not found")
        return obj

    @router.patch("/{item_id}", response_model=read_schema)
    def update_item(item_id: int, payload: update_schema, session: Session = Depends(get_session)):
        obj = service.get(session, item_id)
        if not obj:
            raise HTTPException(404, "Not found")
        if hasattr(_hooks, "pre_update") and callable(_hooks.pre_update):
            _hooks.pre_update(payload, session, obj)
        try:
            return service.update(session, item_id, payload)
        except ValueError:
            raise HTTPException(404, "Not found")

    @router.delete("/{item_id}", status_code=204)
    def delete_item(item_id: int, session: Session = Depends(get_session)):
        obj = service.get(session, item_id)
        if not obj:
            raise HTTPException(404, "Not found")
        if hasattr(_hooks, "pre_delete") and callable(_hooks.pre_delete):
            _hooks.pre_delete(session, obj)
        try:
            service.delete(session, item_id)
        except ValueError:
            raise HTTPException(404, "Not found")

    return router



# app/services/base.py
from typing import Generic, TypeVar, List, Any, Optional
from sqlmodel import Session, SQLModel
from repository.base import Repository, ModelT, CreateT, UpdateT

class Service(Generic[ModelT, CreateT, UpdateT]):
    def __init__(self, repo: Repository[ModelT, CreateT, UpdateT]):
        self.repo = repo

    def get(self, session: Session, id: Any) -> Optional[ModelT]:
        return self.repo.get(session, id)

    def list(self, session: Session, offset: int = 0, limit: int = 100) -> List[ModelT]:
        return self.repo.list(session, offset, limit)

    def create(self, session: Session, data: CreateT) -> ModelT:
        
        return self.repo.create(session, data)

    def update(self, session: Session, id: Any, data: UpdateT) -> ModelT:
        obj = self.repo.get(session, id)
        if not obj:
            raise ValueError("Not found")
        return self.repo.update(session, obj, data)

    def delete(self, session: Session, id: Any) -> None:
        obj = self.repo.get(session, id)
        if not obj:
            raise ValueError("Not found")
        return self.repo.delete(session, obj)



from fastapi import FastAPI
from util.database import init_db
from controller.hero import router as heroes_router
from controller.team import router as teams_router

app = FastAPI(title="FastAPI + SQLModel - MVC + Repository")

init_db()

app.include_router(heroes_router)
app.include_router(teams_router)

@app.get("/")
def health():
    return {"status": "ok"}



from model.models import HeroBase, TeamBase
from typing import  List, Optional
from sqlmodel import Field



class TeamCreate(TeamBase):
    pass

class TeamPublic(TeamBase):
    id: int
    model_config = {"from_attributes": True}

class TeamWithHeroes(TeamPublic):
    heroes: List["HeroPublic"] = []
    model_config = {"from_attributes": True}

class TeamCreate(TeamBase):
    pass

class TeamRead(TeamBase):
    id: int

class TeamUpdate(TeamBase):
    name: Optional[str] = None


class HeroCreate(HeroBase):
    team_id: Optional[int] = None  # permite já criar vinculado a um time

class HeroUpdate(HeroBase):
    name: Optional[str] = Field(default=None, min_length=2, max_length=120)
    secret_name: Optional[str] = None
    age: Optional[int] = Field(default=None, ge=0, le=200)
    
class HeroPublic(HeroBase):
    id: int
    model_config = {"from_attributes": True}

class HeroRead(HeroBase):
    id: int
    team_id: Optional[int] = None



# app/database.py
from sqlmodel import SQLModel, Session, create_engine
from typing import Annotated
from fastapi import Depends
import os

DATABASE_URL = os.getenv("DATABASE_URL", "sqlite:///./app.db")
connect_args = {"check_same_thread": False} if DATABASE_URL.startswith("sqlite") else {}
engine = create_engine(DATABASE_URL, echo=False, connect_args=connect_args)

def init_db() -> None:
    SQLModel.metadata.create_all(engine)

def get_session():
    with Session(engine) as session:
        yield session

SessionDep = Annotated[Session, Depends(get_session)]



# app/routers/hero.py
from fastapi import HTTPException
from sqlmodel import Session, select
from controller.generic import create_crud_router, Hooks
from model.models import Hero, Team
from model.dto import HeroCreate, HeroUpdate, HeroRead

class HeroHooks(Hooks[Hero, HeroCreate, HeroUpdate]):
    def pre_create(self, payload: HeroCreate, session: Session) -> None:
        # se veio team_id, valida
        if payload.team_id is not None and payload.team_id != 0:
            if not session.get(Team, payload.team_id):
                raise HTTPException(400, "team_id inválido")

    def pre_update(self, payload: HeroUpdate, session: Session, obj: Hero) -> None:
        # se vai alterar team_id, valida
        if payload.team_id is not None:
            if payload.team_id != 0 and not session.get(Team, payload.team_id):
                raise HTTPException(400, "team_id inválido")

router = create_crud_router(
    model=Hero,
    create_schema=HeroCreate,
    update_schema=HeroUpdate,
    read_schema=HeroRead,
    prefix="/heroes",
    tags=["heroes"],
    hooks=HeroHooks(),
)



# app/models.py
from typing import Optional, List
from sqlmodel import SQLModel, Field, Relationship

# ---------- TEAM ----------
class TeamBase(SQLModel):
    name: str = Field(min_length=2, max_length=120)

class Team(TeamBase, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    # back_populates liga Team <-> Hero
    heroes: List["Hero"] = Relationship(back_populates="team")


# ---------- HERO ----------
class HeroBase(SQLModel):
    name: str = Field(min_length=2, max_length=120)
    secret_name: Optional[str] = None
    age: Optional[int] = Field(default=None, ge=0, le=200)

class Hero(HeroBase, table=True):
    id: Optional[int] = Field(default=None, primary_key=True, index=True)
    team_id: Optional[int] = Field(default=None, foreign_key="team.id")
    team: Optional[Team] = Relationship(back_populates="heroes")



from .controller.hero import router as HeroRouter
from .controller.team import router as TeamRouter
from .util.database import init_db, get_session

__all__ = [
    "HeroRouter",
    "TeamRouter",
    "init_db",
    "get_session",
]
